import argparse
import os
import tempfile

from cnrclient.commands.command_base import CommandBase
from cnrclient.commands.pull import PullCmd
from cnrclient.plugins.helm import Helm

LOCAL_DIR = os.path.dirname(__file__)


class HelmCmd(CommandBase):
    name = 'helm'
    help_message = 'Deploy with Helm on Kubernetes'
    parse_unknown = True

    def __init__(self, options):
        super(HelmCmd, self).__init__(options)
        self.status = {}

    def exec_helm_cmd(self, cmd, options, helm_opts):
        pull_cmd = PullCmd(options)
        pull_cmd.exec_cmd(render=False)
        helm_cli = Helm()
        output = helm_cli.action(cmd, pull_cmd.path, helm_opts)
        self.status = {'result': output}
        self.render()

    @classmethod
    def _install(cls, options, unknown=None):
        cmd = cls(options)
        cmd.exec_helm_cmd('install', options, unknown)

    @classmethod
    def _upgrade(cls, options, unknown=None):
        cmd = cls(options)
        cmd.exec_helm_cmd('upgrade', options, unknown)

    @classmethod
    def _dep_pull(cls, options, unknown=None):
        cmd = cls(options)
        helm_cli = Helm()
        cmd.status = {'result': helm_cli.build_dep(dest=options.dest, overwrite=options.overwrite)}
        cmd.render()

    @classmethod
    def _init_args(cls, subcmd):
        cls._add_registryhost_option(subcmd)
        cls._add_packagename_option(subcmd)
        cls._add_packageversion_option(subcmd)
        subcmd.add_argument('-t', '--media-type', default='helm', help=argparse.SUPPRESS)

        subcmd.add_argument('--dest', default=tempfile.gettempdir(),
                            help='directory used to extract resources')
        subcmd.add_argument('--tarball', action='store_true', default=True, help=argparse.SUPPRESS)

    @classmethod
    def _init_dep_args(cls, subcmd):
        subcmd.add_argument('--dest', default="appr_charts",
                            help='directory used to extract resources')
        subcmd.add_argument('--overwrite', action='store_true', default=False,
                            help="auto-merge requirements.yaml with the appr dependencies")

    @classmethod
    def _add_arguments(cls, parser):
        sub = parser.add_subparsers()
        install_cmd = sub.add_parser('install')
        upgrade_cmd = sub.add_parser('upgrade')
        dep_pull_cmd = sub.add_parser('dep')
        cls._init_dep_args(dep_pull_cmd)
        cls._init_args(install_cmd)
        cls._init_args(upgrade_cmd)
        install_cmd.set_defaults(func=cls._install)
        upgrade_cmd.set_defaults(func=cls._upgrade)
        dep_pull_cmd.set_defaults(func=cls._dep_pull)

    def _render_dict(self):
        return self.status

    def _render_console(self):
        return self.status['result']



import argparse
import os
import base64
import tempfile
import requests

from cnrclient.pack import pack_kub
from cnrclient.utils import package_filename
from cnrclient.commands.command_base import CommandBase, PackageSplit
from cnrclient.formats.helm.manifest_chart import ManifestChart
from cnrclient.formats import detect_format


class PushCmd(CommandBase):
    name = 'push'
    help_message = "push a package to the registry"
    default_media_type = None

    def __init__(self, options):
        super(PushCmd, self).__init__(options)
        self.registry_host = options.registry_host
        self.force = options.force
        self.manifest = None
        self.media_type = options.media_type
        self.channel = options.channel
        self.version = options.version
        self.filter_files = True
        self.metadata = None
        self.prefix = None
        self.manifest_name = None
        self.package_name = options.package
        self.package_parts = options.package_parts
        self.pname = self.package_parts.get('package', None)
        self.namespace = self.package_parts.get('namespace', None)
        if self.namespace is None:
            self.namespace = options.ns
        self.version_parts = options.version_parts
        if self.version == "default":
            self.version = None
        self.status = ''

    @classmethod
    def _add_arguments(cls, parser):
        cls._add_registryhost_option(parser)
        cls._add_mediatype_option(parser, cls.default_media_type, required=False)
        cls._add_packageversion_option(parser)
        parser.add_argument("--ns", "--namespace", default=None, help=argparse.SUPPRESS)
        parser.add_argument("-f", "--force", action='store_true', default=False, help="force push")
        parser.add_argument("-c", "--channel", default=None, help="Set a channel")
        parser.add_argument("--version-parts", default={}, help=argparse.SUPPRESS)
        parser.add_argument("--package-parts", default={}, help=argparse.SUPPRESS)

        parser.add_argument('package', nargs='?', default=None, action=PackageSplit,
                            help="repository dest")

    def _push(self):
        client = self.RegistryClient(self.registry_host)
        filename = package_filename(self.package_name, self.version, self.media_type)
        kubepath = os.path.join(tempfile.gettempdir(), filename + ".tar.gz")
        pack_kub(kubepath, filter_files=self.filter_files, prefix=self.prefix)
        kubefile = open(kubepath, 'rb')
        body = {
            "manifest_name": self.manifest_name,
            "name": self.package_name,
            "release": self.version,
            "metadata": self.metadata,
            "media_type": self.media_type,
            "blob": base64.b64encode(kubefile.read())
        }
        try:
            client.push(self.package_name, body, self.force)
            self.status = "package: %s (%s | %s) pushed\n" % (self.package_name, self.version,
                                                              self.media_type)
        except requests.exceptions.RequestException as exc:
            if not (self.channel and exc.response.status_code in [409, 400]):
                raise
        kubefile.close()
        os.remove(kubepath)
        if self.channel:
            client.create_channel_release(self.package_name, self.channel, self.version)
            self.status += ">>> Release '%s' added to '%s'" % (self.version, self.channel)

    def _chart(self):
        self.manifest = ManifestChart()
        self.manifest_name = self.manifest.name
        if self.pname is None:
            self.pname = self.manifest.name
        self.prefix = self.pname
        self.filter_files = False
        if self.namespace is None:
            raise argparse.ArgumentTypeError("Missing option: --namespace")
        self.package_name = "%s/%s" % (self.namespace, self.pname)
        if self.version is None:
            self.version = self.manifest.version
        self.metadata = self.manifest.metadata()

    def _all_formats(self):
        self.filter_files = False
        if self.version is None or self.version == "default":
            raise argparse.ArgumentTypeError("Missing option: --version")
        if self.package_name is None:
            raise argparse.ArgumentTypeError("Missing option: --name")
        self.namespace, self.pname = self.package_name.split("/")

    def _kpm(self):
        raise NotImplementedError

    def _init(self):
        if self.media_type is None:
            self.media_type = detect_format(".").media_type
        if self.media_type in ["kpm", "kpm-compose"]:
            self._kpm()
        elif self.media_type in ['helm', 'chart']:
            self._chart()
        else:
            self._all_formats()

    def _call(self):
        self._init()
        self._push()

    def _render_dict(self):
        return {
            "manifest_name": self.manifest_name,
            "package": self.package_name,
            "version": self.version,
            "media_type": self.media_type,
            "channel": self.channel
        }

    def _render_console(self):
        return self.status



from __future__ import print_function
import os
import sys
import subprocess

import yaml

from cnrclient.client import CnrClient
from cnrclient.utils import mkdir_p, parse_package_name
from cnrclient.pack import CnrPackage

DEFAULT_CHARTS = "appr_charts"


def parse_version(version):
    """
     Converts a version string to a dict with following rules:
       if string starts with ':' it is a channel
       if string starts with 'sha256' it is a digest
       else it is a release
    """
    if version[0] == ':' or version.startswith('channel:'):
        parts = {'key': 'channel', 'value': version.split(':')[1]}
    elif version.startswith('sha256:'):
        parts = {'key': 'digest', 'value': version.split('sha256:')[1]}
    else:
        parts = {'key': 'version', 'value': version}

    return parts


class Helm(object):
    def __init__(self):
        pass

    def download_appr_deps(self, deps, dest=DEFAULT_CHARTS, tarball=False):
        """
            Creates a directory 'dep_charts' to download and extract all dependencies
            fetched from the app-registry server.
            returns a helm dependency list
        """
        mkdir_p(dest)
        helm_deps = {}
        for dep in deps:
            package_parts = parse_package_name(dep['name'])
            name = package_parts['package']

            vparts = parse_version(dep['version'])
            client = CnrClient(package_parts['host'])
            package_name = '%s/%s' % (package_parts['namespace'], name)

            pullpack = client.pull_json(package_name, version_parts=vparts, media_type='helm')
            package = CnrPackage(pullpack['blob'], b64_encoded=True)
            release = pullpack['release']
            packagepath = os.path.join(dest, package_parts['namespace'])
            print('Pulled package: %s (%s) \n -> %s' % (dep['name'], release, packagepath),
                  file=sys.stderr)
            if tarball:
                with open('%s-%s.tgz' % (name, release), 'wb') as tarfile:
                    tarfile.write(package.blob)
            package.extract(packagepath)

            helm_deps[name] = {
                'name': name,
                'version': release,
                'repository': 'file://%s/%s' % (packagepath, name)
            }
        return helm_deps

    def build_dep(self, dest=DEFAULT_CHARTS, overwrite=False):
        """
            Reads the dependencies from the requirements.yaml, downloads the packages and updates
            the requirements.yaml.
            Returns status
        """
        with open('requirements.yaml', 'rb') as requirefile:
            deps = yaml.safe_load(requirefile.read())
        helm_deps = {}
        if 'appr' in deps and deps['appr']:
            helm_deps = self.download_appr_deps(deps['appr'], dest)
            dict_deps = {dep['name']: dep for dep in deps['dependencies']}
            dict_deps.update(helm_deps)
            deps['dependencies'] = dict_deps.values()
            requirement_output = yaml.safe_dump(deps, default_flow_style=False)
            if overwrite:
                with open('requirements.yaml', 'wb') as requirefile:
                    requirefile.write(requirement_output)
                return 'Updated requirements.yaml'
            else:
                return requirement_output
        else:
            return "No appr-registries dependencies"

    def action(self, cmd, package_path, helm_opts=None):
        cmd = [cmd]
        if helm_opts:
            cmd = cmd + helm_opts
        cmd.append(package_path)
        return self._call(cmd)

    def _call(self, cmd):
        command = ['helm'] + cmd
        try:
            return subprocess.check_output(command, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as exc:
            return exc.output
